/*
 * The MIT License
 * 
 * Copyright (c) 2004-2009, Sun Microsystems, Inc., Kohsuke Kawaguchi
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package hudson.remoting.jnlp;

import javax.swing.*;
import java.awt.event.KeyEvent;

/**
 * Main menu of {@link MainDialog}.
 *
 * @author Kohsuke Kawaguchi
 */
public final class MainMenu extends JMenuBar {
    private final MainDialog owner;
    private JMenu fileMenu;

    MainMenu(MainDialog owner) {
        this.owner = owner;
    }

    /**
     * Obtains the file menu (and creates it if necessary),
     * so that the caller can add items in this menu.
     */
    public JMenu getFileMenu() {
        if(fileMenu==null) {
            fileMenu = new JMenu("File");
            fileMenu.setMnemonic(KeyEvent.VK_F);
            add(fileMenu,0);
        }
        return fileMenu;
    }

    /**
     * Reflects the changes made in the menu objects to GUI.
     */
    public void commit() {
        invalidate();
        repaint();
        if(getComponentCount()>0) {
            owner.setJMenuBar(this);
            // work around for paint problem. see http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4949810
            if(owner.isVisible())
                owner.setVisible(true);
        }
    }
}
